@@ -10,7 +10,7 @@ from .float import avoid_collisions, float_layout, get_clearance
 from .grid import grid_layout
 from .inline import iter_line_boxes
 from .min_max import handle_min_max_width
-from .percent import resolve_percentages, resolve_position_percentages
+from .percent import percentage, resolve_percentages, resolve_position_percentages
 from .replaced import block_replaced_box_layout
 from .table import table_layout, table_wrapper_width
 
@@ -81,7 +81,7 @@ def block_level_layout_switch(context, box, bottom_space, skip_stack,
     elif isinstance(box, boxes.FlexBox):
         result = flex_layout(
             context, box, bottom_space, skip_stack, containing_block,
-            page_is_empty, absolute_boxes, fixed_boxes)
+            page_is_empty, absolute_boxes, fixed_boxes, discard)
     elif isinstance(box, boxes.GridBox):
         result = grid_layout(
             context, box, bottom_space, skip_stack, containing_block,
@@ -124,8 +124,11 @@ def block_box_layout(context, box, bottom_space, skip_stack,
     result = block_container_layout(
         context, box, bottom_space, skip_stack, page_is_empty,
         absolute_boxes, fixed_boxes, adjoining_margins, discard, max_lines)
-    new_box = result[0]
-    if new_box and new_box.is_table_wrapper:
+    # TODO: columns and flex items shouldn't be block boxes, this condition
+    # would then be useless when this is fixed.
+    if not (new_box := result[0]) or new_box.is_column or new_box.is_flex_item:
+        return result
+    if new_box.is_table_wrapper or new_box.establishes_formatting_context():
         # Don't collide with floats
         # https://www.w3.org/TR/CSS21/visuren.html#floats
         position_x, position_y, _ = avoid_collisions(
@@ -275,6 +278,7 @@ def _out_of_flow_layout(context, box, index, child, new_children,
             last_in_flow_child = find_last_in_flow_child(new_children)
             page_break = block_level_page_break(last_in_flow_child, child)
             resume_at = {index: None}
+            out_of_flow_resume_at = None
             stop = True
             if new_children and avoid_page_break(page_break, context):
                 # Can’t break inside float, find an earlier page break.
@@ -283,7 +287,7 @@ def _out_of_flow_layout(context, box, index, child, new_children,
                 if result:
                     # Earlier page break found, drop whole child rendering.
                     new_children[:], resume_at = result
-                    new_child = out_of_flow_resume_at = None
+                    new_child = None
 
     # Running element layout.
     elif child.is_running():
@@ -294,24 +298,25 @@ def _out_of_flow_layout(context, box, index, child, new_children,
     return stop, resume_at, new_child, out_of_flow_resume_at
 
 
-def _break_line(context, box, line, new_children, lines_iterator,
-                page_is_empty, index, skip_stack, resume_at, absolute_boxes,
-                fixed_boxes):
+def _break_line(context, box, line, new_children, next_lines, page_is_empty, index,
+                skip_stack, resume_at, absolute_boxes, fixed_boxes):
+    """Break line where allowed by orphans and widows.
+
+    Return (abort, stop, resume_at).
+
+    """
     over_orphans = len(new_children) - box.style['orphans']
     if over_orphans < 0 and not page_is_empty:
         # Reached the bottom of the page before we had
         # enough lines for orphans, cancel the whole box.
+        remove_placeholders(context, line.children, absolute_boxes, fixed_boxes)
         return True, False, resume_at
     # How many lines we need on the next page to satisfy widows
     # -1 for the current line.
-    needed = box.style['widows'] - 1
-    if needed:
-        for _ in lines_iterator:
-            needed -= 1
-            if needed == 0:
-                break
+    needed = max(box.style['widows'] - 1 - next_lines, 0)
     if needed > over_orphans and not page_is_empty:
         # Total number of lines < orphans + widows
+        remove_placeholders(context, line.children, absolute_boxes, fixed_boxes)
         return True, False, resume_at
     if needed and needed <= over_orphans:
         # Remove lines to keep them for the next page
@@ -360,24 +365,52 @@ def _linebox_layout(context, box, index, child, new_children, page_is_empty,
         else:
             offset_y = 0
 
-        # Allow overflow if the first line of the page is higher
-        # than the page itself so that we put *something* on this
-        # page and can advance in the context.
+        # Allow overflow if the first line of the page is higher than the page itself so
+        # that we put *something* on this page and can advance in the context.
         overflow = (
             (new_children or not page_is_empty) and
             context.overflows_page(bottom_space, new_position_y + offset_y))
         if overflow:
-            abort, stop, resume_at = _break_line(
-                context, box, line, new_children, lines_iterator,
-                page_is_empty, index, skip_stack, resume_at, absolute_boxes,
-                fixed_boxes)
+            # If we couldn’t break the line before but can break now, first try to
+            # report footnotes and see if we don’t overflow.
+            could_break_before = can_break_now = True
+            next_lines = len(tuple(lines_iterator))
+            if len(new_children) + 1 < box.style['orphans']:
+                can_break_now = False
+            elif next_lines < box.style['widows']:
+                can_break_now = False
+            if len(new_children) < box.style['orphans']:
+                could_break_before = False
+            elif next_lines + 1 < box.style['widows']:
+                could_break_before = False
+            report = not context.in_column and can_break_now and not could_break_before
+            reported_footnotes = 0
+            while report and context.current_page_footnotes:
+                context.report_footnote(context.current_page_footnotes[-1])
+                reported_footnotes += 1
+                if not context.overflows_page(bottom_space, new_position_y + offset_y):
+                    new_children.append(line)
+                    stop = True
+                    break
+            else:
+                abort, stop, resume_at = _break_line(
+                    context, box, line, new_children, next_lines,
+                    page_is_empty, index, skip_stack, resume_at, absolute_boxes,
+                    fixed_boxes)
+
+            # Revert reported footnotes, as they’ve been reported starting from the last
+            # one.
+            if reported_footnotes >= 2:
+                extra = context.reported_footnotes[-1:-reported_footnotes-1:-1]
+                context.reported_footnotes[-reported_footnotes:] = extra
+
             break
 
         # TODO: this is incomplete.
         # See https://drafts.csswg.org/css-page-3/#allowed-pg-brk
         # "When an unforced page break occurs here, both the adjoining
         #  ‘margin-top’ and ‘margin-bottom’ are set to zero."
-        # See https://github.com/Kozea/WeasyPrint/issues/115
+        # See issue #115.
         elif page_is_empty and context.overflows_page(
                 bottom_space, new_position_y):
             # Remove the top border when a page is empty and the box is
@@ -407,9 +440,10 @@ def _linebox_layout(context, box, index, child, new_children, page_is_empty,
                     # even try.
                     if new_children or not page_is_empty:
                         if footnote.style['footnote_policy'] == 'line':
+                            next_lines = len(tuple(lines_iterator))
                             abort, stop, resume_at = _break_line(
                                 context, box, line, new_children,
-                                lines_iterator, page_is_empty, index,
+                                next_lines, page_is_empty, index,
                                 skip_stack, resume_at, absolute_boxes,
                                 fixed_boxes)
                             break_linebox = True
@@ -490,11 +524,22 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
                 adjoining_margins = []
                 position_y = box.content_box_y()
 
-    if adjoining_margins and box.is_table_wrapper:
-        collapsed_margin = collapse_margin(adjoining_margins)
-        child.position_y += collapsed_margin
-        position_y += collapsed_margin
-        adjoining_margins = []
+    # TODO: Merge this with block_container_layout, block_level_layout, _in_flow_layout,
+    # and check code above.
+    if adjoining_margins:
+        if box.is_table_wrapper:  # should not be a special case
+            collapsed_margin = collapse_margin(adjoining_margins)
+            child.position_y += collapsed_margin
+            adjoining_margins = []
+        elif not isinstance(child, boxes.BlockBox):  # blocks handle that themselves
+            if child.style['margin_top'] == 'auto':
+                margin_top = 0
+            else:
+                margin_top = percentage(child.style['margin_top'], box.width)
+            adjoining_margins.append(margin_top)
+            offset_y = collapse_margin(adjoining_margins) - margin_top
+            child.position_y += offset_y
+            adjoining_margins = []
 
     page_is_empty_with_no_children = page_is_empty and not any(
         child for child in new_children
@@ -509,15 +554,6 @@ def _in_flow_layout(context, box, index, child, new_children, page_is_empty,
          fixed_boxes, adjoining_margins, discard, max_lines)
 
     if new_child is not None:
-        # We need to do this after the child layout to have the
-        # used value for margin_top (eg. it might be a percentage.)
-        if not isinstance(new_child, (boxes.BlockBox, boxes.TableBox)):
-            adjoining_margins.append(new_child.margin_top)
-            offset_y = (
-                collapse_margin(adjoining_margins) - new_child.margin_top)
-            new_child.translate(0, offset_y)
-        # else: blocks handle that themselves.
-
         if not collapsing_through:
             new_content_position_y = (
                 new_child.content_box_y() + new_child.height)
@@ -623,13 +659,12 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                            page_is_empty, absolute_boxes, fixed_boxes,
                            adjoining_margins, discard, max_lines):
     """Set the ``box`` height."""
-    # TODO: boxes.FlexBox is allowed here because flex_layout calls
-    # block_container_layout, there's probably a better solution.
-    assert isinstance(box, (boxes.BlockContainerBox, boxes.FlexBox))
+    assert isinstance(box, boxes.BlockContainerBox)
 
-    if establishes_formatting_context(box):
+    if box.establishes_formatting_context():
         context.create_block_formatting_context()
 
+    # TODO: merge this with _in_flow_layout, flex_layout…
     is_start = skip_stack is None
     box.remove_decoration(start=not is_start, end=False)
 
@@ -649,7 +684,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     collapsing_with_children = not (
         box.border_top_width or box.padding_top or box.is_flex_item or
-        box.is_grid_item or establishes_formatting_context(box) or
+        box.is_grid_item or box.establishes_formatting_context() or
         box.is_for_root_element)
     if collapsing_with_children:
         # Not counting margins in adjoining_margins, if any
@@ -668,8 +703,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     new_children = []
     next_page = {'break': 'any', 'page': None}
-    all_footnotes = []
-    broken_out_of_flow = {}
+    broken_out_of_flow = []
 
     last_in_flow_child = None
 
@@ -695,8 +729,13 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                     absolute_boxes, fixed_boxes, adjoining_margins,
                     bottom_space))
             if out_of_flow_resume_at:
-                broken_out_of_flow[new_child] = (
+                context.broken_out_of_flow[new_child] = (
                     child, box, out_of_flow_resume_at)
+                broken_out_of_flow.append(new_child)
+            if child.is_outside_marker:
+                new_child.position_x = box.border_box_x()
+                if child.style['direction'] == 'rtl':
+                    new_child.position_x += box.width + box.padding_right
 
         elif isinstance(child, boxes.LineBox):
             (abort, stop, resume_at, position_y,
@@ -707,7 +746,6 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                 draw_bottom_decoration, max_lines)
             draw_bottom_decoration |= resume_at is None
             adjoining_margins = []
-            all_footnotes += new_footnotes
 
         else:
             (abort, stop, resume_at, position_y, adjoining_margins,
@@ -747,7 +785,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                 max_lines)
         elif stop:
             if box.height != 'auto':
-                if context.overflows(box.position_y + box.height, position_y):
+                if context.overflows(box.position_y + box.border_height(), position_y):
                     # Box heigh is fixed and it doesn’t overflow page, forget
                     # overflowing children.
                     resume_at = None
@@ -757,20 +795,18 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     else:
         resume_at = None
 
-    box_is_fragmented = resume_at is not None
+    box_is_fragmented = resume_at is not None or box.force_fragmentation
     if box.style['continue'] == 'discard':
         resume_at = None
 
     if (box_is_fragmented and
             avoid_page_break(box.style['break_inside'], context) and
             not page_is_empty):
-        for footnote in all_footnotes:
-            context.unlayout_footnote(footnote)
-        return (
-            None, None, {'break': 'any', 'page': None}, [], False, max_lines)
-
-    for key, value in broken_out_of_flow.items():
-        context.broken_out_of_flow[key] = value
+        remove_placeholders(
+            context, [*new_children, *box.children[skip:]], absolute_boxes, fixed_boxes)
+        for child in broken_out_of_flow:
+            del context.broken_out_of_flow[child]
+        return None, None, {'break': 'any', 'page': None}, [], False, max_lines
 
     if collapsing_with_children:
         box.position_y += (
@@ -798,7 +834,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
 
     if (box.border_bottom_width or
             box.padding_bottom or
-            establishes_formatting_context(box) or
+            box.establishes_formatting_context() or
             box.is_for_root_element or
             box.is_table_wrapper):
         position_y += collapse_margin(adjoining_margins)
@@ -823,7 +859,10 @@ def block_container_layout(context, box, bottom_space, skip_stack,
                 float_box.position_y + float_box.margin_height()
                 for float_box in context.excluded_shapes)
             position_y = max(max_float_position_y, position_y)
-        new_box.height = position_y - new_box.content_box_y()
+        if position_y == new_box.content_box_y() == inf:
+            new_box.height = 0
+        else:
+            new_box.height = position_y - new_box.content_box_y()
 
     if new_box.style['position'] == 'relative':
         # New containing block, resolve the layout of the absolute descendants
@@ -835,7 +874,7 @@ def block_container_layout(context, box, bottom_space, skip_stack,
     for child in new_box.children:
         relative_positioning(child, (new_box.width, new_box.height))
 
-    if establishes_formatting_context(new_box):
+    if new_box.establishes_formatting_context():
         context.finish_block_formatting_context(new_box)
 
     if discard or not box_is_fragmented:
@@ -873,30 +912,6 @@ def collapse_margin(adjoining_margins):
     return max(positives) + min(negatives)
 
 
-def establishes_formatting_context(box):
-    """Return whether a box establishes a block formatting context.
-
-    See https://www.w3.org/TR/CSS2/visuren.html#block-formatting
-
-    """
-    return (
-        box.is_floated()
-    ) or (
-        box.is_absolutely_positioned()
-    ) or (
-        # TODO: columns shouldn't be block boxes, this condition would then be
-        # useless when this is fixed
-        box.is_column
-    ) or (
-        isinstance(box, boxes.BlockContainerBox) and
-        not isinstance(box, boxes.BlockBox)
-    ) or (
-        isinstance(box, boxes.BlockBox) and box.style['overflow'] != 'visible'
-    ) or (
-        'flow-root' in box.style['display']
-    )
-
-
 def block_level_page_break(sibling_before, sibling_after):
     """Get the correct page break value between siblings.
 
