@@ -1,5 +1,7 @@
 """Layout for floating boxes."""
 
+from math import inf
+
 from ..formatting_structure import boxes
 from .min_max import handle_min_max_width
 from .percent import resolve_percentages, resolve_position_percentages
@@ -68,7 +70,7 @@ def float_layout(context, box, containing_block, absolute_boxes, fixed_boxes,
             context, box, bottom_space=bottom_space,
             skip_stack=skip_stack, containing_block=containing_block,
             page_is_empty=True, absolute_boxes=absolute_boxes,
-            fixed_boxes=fixed_boxes)
+            fixed_boxes=fixed_boxes, discard=False)
     elif isinstance(box, boxes.GridContainerBox):
         box, resume_at, _, _, _ = grid_layout(
             context, box, bottom_space=bottom_space,
@@ -115,9 +117,14 @@ def find_float_position(context, box, containing_block):
 
 def get_clearance(context, box, collapsed_margin=0):
     """Return None if there is no clearance, otherwise the clearance value."""
+    # Box should be after shape that’s broken on this page.
+    for broken_shape in context.broken_out_of_flow:
+        if broken_shape.is_floated():
+            if box.style['clear'] in (broken_shape.style['float'], 'both'):
+                return inf
+    # Hypothetical position is the position of the top border edge
     clearance = None
     hypothetical_position = box.position_y + collapsed_margin
-    # Hypothetical position is the position of the top border edge
     for excluded_shape in context.excluded_shapes:
         if box.style['clear'] in (excluded_shape.style['float'], 'both'):
             y, h = excluded_shape.position_y, excluded_shape.margin_height()
@@ -179,12 +186,12 @@ def avoid_collisions(context, box, containing_block, outer=True):
             # Points 3, 7 and 8
             if box_width > max_right_bound - max_left_bound:
                 # The box does not fit here
-                new_positon_y = min(
+                new_position_y = min(
                     shape.position_y + shape.margin_height()
                     for shape in colliding_shapes)
-                if new_positon_y > position_y:
+                if new_position_y > position_y:
                     # We can find a solution with a higher position_y
-                    position_y = new_positon_y
+                    position_y = new_position_y
                     continue
                 # No solution, we must put the box here
         break
@@ -195,12 +202,13 @@ def avoid_collisions(context, box, containing_block, outer=True):
     # - line boxes
     # - table wrappers
     # - block-level replaced box
-    # - element establishing new formatting contexts (not handled)
+    # - element establishing new formatting contexts
     assert (
         (box.style['float'] in ('right', 'left')) or
         isinstance(box, boxes.LineBox) or
         box.is_table_wrapper or
-        isinstance(box, boxes.BlockReplacedBox))
+        isinstance(box, boxes.BlockReplacedBox) or
+        box.establishes_formatting_context())
 
     # The x-position of the box depends on its type.
     position_x = max_left_bound
@@ -215,9 +223,11 @@ def avoid_collisions(context, box, containing_block, outer=True):
                 # bound.
                 position_x = max_right_bound - box_width
             else:
-                # The position of the right border of the replaced box is at
-                # the right bound.
-                assert isinstance(box, boxes.BlockReplacedBox)
+                # The position of the right border of the replaced box or
+                # formatting context is at the right bound.
+                assert (
+                    isinstance(box, boxes.BlockReplacedBox) or
+                    box.establishes_formatting_context())
                 position_x = max_right_bound - box_width
 
     available_width = max_right_bound - max_left_bound
